# flake8: noqa

try:
    from ._version import version as __version__
except ImportError:
    __version__ = ''

# Import all submodules.
from . import aperture
from . import atmosphere
from . import config
from . import coronagraphy
from . import field
from . import fourier
from . import interpolation
from . import metrics
from . import mode_basis
from . import optics
from . import plotting
from . import propagation
from . import util
from . import wavefront_control
from . import wavefront_sensing

# Import all core submodules in default namespace.
from .aperture import *
from .atmosphere import *
from .config import *
from .coronagraphy import *
from .field import *
from .fourier import *
from .interpolation import *
from .metrics import *
from .mode_basis import *
from .optics import *
from .plotting import *
from .propagation import *
from .util import *
from .wavefront_control import *
from .wavefront_sensing import *

# Export default namespaces.
__all__ = []
__all__.extend(aperture.__all__)
__all__.extend(atmosphere.__all__)
__all__.extend(config.__all__)
__all__.extend(coronagraphy.__all__)
__all__.extend(field.__all__)
__all__.extend(fourier.__all__)
__all__.extend(interpolation.__all__)
__all__.extend(metrics.__all__)
__all__.extend(mode_basis.__all__)
__all__.extend(optics.__all__)
__all__.extend(plotting.__all__)
__all__.extend(propagation.__all__)
__all__.extend(util.__all__)
__all__.extend(wavefront_control.__all__)
__all__.extend(wavefront_sensing.__all__)



import numpy as np
from scipy.linalg import blas
from .fourier_transform import FourierTransform, multiplex_for_tensor_fields, _get_float_and_complex_dtype
from ..field import Field
from ..config import Configuration
import numexpr as ne

class MatrixFourierTransform(FourierTransform):
    '''A Matrix Fourier Transform (MFT) object.

    This Fourier transform is based on the MFT described in [Soummer2007]_. It requires both
    the input and output grid to be separated in Cartesian coordinates. Additionally, due to
    current implementation limitations, this Fourier transform only supports one- and two-dimensional
    grids.

    .. [Soummer2007] Soummer et al. 2007, "Fast computation of Lyot-style
        coronagraph propagation".

    Parameters
    ----------
    input_grid : Grid
        The grid that is expected for the input field.
    output_grid : Grid
        The grid that is produced by the Fourier transform.
    precompute_matrices : boolean or None
        Whether to precompute the matrices used in the MFT. Turning this on will provide a 20-30%
        speedup, in exchange for higher memory usage.If this is False, the matrices will be
        calculated each time a Fourier transform is performed. If this is True, the matrices will be
        calculated once, and reused for future evaluations. If this is None, the choice will be
        determined by the configuration file.
    allocate_intermediate : boolean or None
        Whether to reserve memory for the intermediate result for the MFT. This provides a 5-10%
        speedup in exchange for higher memory usage. If this is None, the choice will be determined
        by the configuration file.

    Raises
    ------
    ValueError
        If the input grid is not separated in Cartesian coordinates, if it's not one- or two-
        dimensional, or if the output grid has a different dimension than the input grid.
    '''
    def __init__(self, input_grid, output_grid, precompute_matrices=None, allocate_intermediate=None):
        # Check input grid assumptions
        if not input_grid.is_separated or not input_grid.is_('cartesian'):
            raise ValueError('The input_grid must be separable in cartesian coordinates.')
        if not output_grid.is_separated or not output_grid.is_('cartesian'):
            raise ValueError('The output_grid must be separable in cartesian coordinates.')
        if input_grid.ndim not in [1, 2]:
            raise ValueError('The input_grid must be one- or two-dimensional.')
        if input_grid.ndim != output_grid.ndim:
            raise ValueError('The input_grid must have the same dimensions as the output_grid.')

        self.input_grid = input_grid
        self.output_grid = output_grid

        self.shape_input = input_grid.shape
        self.shape_output = output_grid.shape

        self.ndim = input_grid.ndim

        # Get the value from the configuration file if left at default.
        if precompute_matrices is None:
            precompute_matrices = Configuration().fourier.mft.precompute_matrices
        self.precompute_matrices = precompute_matrices

        # Get the value from the configuration file if left at default.
        if allocate_intermediate is None:
            allocate_intermediate = Configuration().fourier.mft.allocate_intermediate
        self.allocate_intermediate = allocate_intermediate

        self.matrices_dtype = None
        self.intermediate_dtype = None
        self._remove_matrices()

    def _compute_matrices(self, dtype):
        '''Compute the matrices for the MFT using the specified data type.

        Parameters
        ---
        dtype : numpy data type
            The data type for which to calculate the matrices.
        '''
        # Set the correct complex and real data type, based on the input data type.
        float_dtype, complex_dtype = _get_float_and_complex_dtype(dtype)

        # Check if the matrices need to be (re)calculated.
        if self.matrices_dtype != complex_dtype:
            self.weights_input = (self.input_grid.weights).astype(float_dtype, copy=False)
            self.weights_output = (self.output_grid.weights / (2 * np.pi)**self.ndim).astype(float_dtype, copy=False)

            # If all input weights are all the same, use a scalar instead.
            if not np.isscalar(self.weights_input) and np.all(self.weights_input == self.weights_input[0]):
                self.weights_input = self.weights_input[0]

            # If all output weights are all the same, use a scalar instead.
            if not np.isscalar(self.weights_output) and np.all(self.weights_output == self.weights_output[0]):
                self.weights_output = self.weights_output[0]

            if self.ndim == 1:
                xu = np.outer(self.output_grid.x, self.input_grid.x)
                self.M = ne.evaluate('exp(-1j * xu)', local_dict={'xu': xu}).astype(complex_dtype, copy=False)
            elif self.ndim == 2:
                x, y = self.input_grid.coords.separated_coords
                u, v = self.output_grid.coords.separated_coords

                vy = np.outer(v, y)
                xu = np.outer(x, u)

                self.M1 = ne.evaluate('exp(-1j * vy)', local_dict={'vy': vy}).astype(complex_dtype, copy=False)
                self.M2 = ne.evaluate('exp(-1j * xu)', local_dict={'xu': xu}).astype(complex_dtype, copy=False)

            self.matrices_dtype = complex_dtype

        # Checki if the intermediate array needs to be (re)allocated.
        if self.intermediate_dtype != complex_dtype:
            if self.ndim == 2:
                self.intermediate_array = np.empty((self.input_grid.shape[0], self.M2.shape[1]), dtype=complex_dtype)

                self.intermediate_dtype = complex_dtype

    def _remove_matrices(self):
        '''Remove the matrices after a Fourier transform.

        This is is used to clean up the used matrices after a Fourier transform operation.
        '''
        if not self.precompute_matrices:
            if self.ndim == 1:
                self.M = None
            elif self.ndim == 2:
                self.M1 = None
                self.M2 = None

            self.matrices_dtype = None

        if not self.allocate_intermediate:
            if self.ndim == 2:
                self.intermediate_array = None
                self.intermediate_dtype = None

    @multiplex_for_tensor_fields
    def forward(self, field):
        '''Returns the forward Fourier transform of the :class:`Field` field.

        Parameters
        ----------
        field : Field
            The field to Fourier transform.

        Returns
        --------
        Field
            The Fourier transform of the field.
        '''
        self._compute_matrices(field.dtype)
        field = field.astype(self.matrices_dtype, copy=False)

        if self.ndim == 1:
            f = field * self.weights_input
            res = np.dot(self.M, f)
        elif self.ndim == 2:
            # Use handcoded BLAS call. BLAS is better when all inputs are Fortran ordered,
            # so we apply matrix multiplications on the transpose of each of the arrays
            # (which are C ordered).
            if field.dtype == 'complex64':
                gemm = blas.cgemm
            else:
                gemm = blas.zgemm

            if np.isscalar(self.weights_input):
                # Weights can be included in the gemm call as that multiplication
                # happens anyway (and it saves an array copy).
                f = field.reshape(self.shape_input)
                alpha = self.weights_input
            else:
                # Fallback in case the weights is not a scalar.
                f = (field * self.weights_input).reshape(self.shape_input)
                alpha = 1

            gemm(1, self.M2.T, f.T, c=self.intermediate_array.T, overwrite_c=True)
            res = gemm(alpha, self.intermediate_array.T, self.M1.T).T.reshape(-1)

        self._remove_matrices()

        return Field(res, self.output_grid)

    @multiplex_for_tensor_fields
    def backward(self, field):
        '''Returns the inverse Fourier transform of the :class:`Field` field.

        Parameters
        ----------
        field : Field
            The field to inverse Fourier transform.

        Returns
        --------
        Field
            The inverse Fourier transform of the field.
        '''
        self._compute_matrices(field.dtype)
        field = field.astype(self.matrices_dtype, copy=False)

        if self.ndim == 1:
            f = field * self.weights_output
            res = np.dot(self.M.conj().T, f)
        elif self.ndim == 2:
            # Use handcoded BLAS call. BLAS is better when all inputs are Fortran ordered,
            # so we apply matrix multiplications on the transpose of each of the arrays
            # (which are C ordered). Adjoint is handled by GEMM, which avoids an array
            # copy for these array as well.
            if field.dtype == 'complex64':
                gemm = blas.cgemm
            else:
                gemm = blas.zgemm

            if np.isscalar(self.weights_output):
                # Weights can be included in the gemm call as that multiplication
                # happens anyway (and it saves an array copy).
                f = field.reshape(self.shape_output)
                alpha = self.weights_output
            else:
                # Fallback in case the weights is not a scalar.
                f = (field * self.weights_output).reshape(self.shape_output)
                alpha = 1

            # Use trans_a=2 and trans_b=2 to apply the conjugte transpose on the a and b arrays.
            gemm(1, f.T, self.M1.T, trans_b=2, c=self.intermediate_array.T, overwrite_c=True)
            res = gemm(alpha, self.M2.T, self.intermediate_array.T, trans_a=2).T.reshape(-1)

        self._remove_matrices()

        return Field(res, self.input_grid)



__all__ = [
    'generate_app_keller',
    'generate_app_por',
    'VectorApodizingPhasePlate',
    'KnifeEdgeLyotCoronagraph',
    'LyotCoronagraph',
    'MultiScaleCoronagraph',
    'OccultedLyotCoronagraph',
    'PerfectCoronagraph',
    'FQPMCoronagraph',
    'VortexCoronagraph',
    'VectorVortexCoronagraph',
    'make_ravc_masks',
    'get_ravc_planet_transmission'
]

from .apodizing_phase_plate import *
from .fqpm import *
from .knife_edge import *
from .lyot import *
from .multi_scale import *
from .perfect_coronagraph import *
from .shaped_pupil import *
from .vortex import *



from hcipy import *
import numpy as np
import copy
import contextlib
import pytest
import pickle

def test_field_dot():
    grid = make_pupil_grid(2)

    a = np.random.randn(3, grid.size)
    A = np.random.randn(3, 3, grid.size)

    a = Field(a, grid)
    A = Field(A, grid)

    b = field_dot(A, a)
    bb = np.array([A[..., i].dot(a[..., i]) for i in range(grid.size)]).T

    assert np.allclose(b, bb)

    b = field_dot(a, a)
    bb = np.array([a[..., i].dot(a[..., i]) for i in range(grid.size)]).T

    assert np.allclose(b, bb)

    B = field_dot(A, A)
    BB = np.empty_like(B)
    for i in range(grid.size):
        BB[..., i] = A[..., i].dot(A[..., i])

    assert np.allclose(B, BB)

    b = field_dot(a, a)
    bb = np.array([a[..., i].dot(a[..., i]) for i in range(grid.size)])

    assert np.allclose(b, bb)

    n = np.random.randn(3)

    b = field_dot(A, n)
    bb = np.array([A[..., i].dot(n) for i in range(grid.size)]).T

    assert np.allclose(b, bb)

    b = field_dot(n, A)
    bb = np.array([n.dot(A[..., i]) for i in range(grid.size)]).T

    assert np.allclose(b, bb)

    N = np.random.randn(3, 3)

    B = field_dot(A, N)
    BB = np.empty_like(B)
    for i in range(grid.size):
        BB[..., i] = A[..., i].dot(N)

    assert np.allclose(B, BB)

def test_field_trace():
    grid = make_pupil_grid(2)

    A = Field(np.random.randn(3, 3, grid.size), grid)

    B = field_trace(A)
    BB = np.array([np.trace(A[..., i]) for i in range(grid.size)])

    assert np.allclose(B, BB)

def test_field_inv():
    grid = make_pupil_grid(2)

    A = Field(np.random.randn(3, 3, grid.size), grid)

    B = field_inv(A)
    BB = np.empty_like(B)
    for i in range(grid.size):
        BB[..., i] = np.linalg.inv(A[..., i])

    assert np.allclose(B, BB)

def test_field_transpose():
    grid = make_pupil_grid(2)

    A = Field(np.random.randn(3, 3, grid.size), grid)

    B = field_transpose(A)
    BB = np.empty_like(B)
    for i in range(grid.size):
        BB[..., i] = A[..., i].T

    assert np.allclose(B, BB)

def test_field_conjugate_transpose():
    grid = make_pupil_grid(2)

    A = Field(np.random.randn(3, 3, grid.size), grid)

    B = field_conjugate_transpose(A)
    BB = np.empty_like(B)
    for i in range(grid.size):
        BB[..., i] = A[..., i].T.conj()

    assert np.allclose(B, BB)

def test_field_adjoint():
    grid = make_pupil_grid(2)

    A = Field(np.random.randn(3, 3, grid.size), grid)

    B = field_adjoint(A)
    BB = np.empty_like(B)
    for i in range(grid.size):
        BB[..., i] = np.linalg.inv(A[..., i]) * np.linalg.det(A[..., i])

    assert np.allclose(B, BB)

def test_field_inverse_tikhonov():
    grid = make_pupil_grid(2)

    A = Field(np.random.randn(3, 3, grid.size), grid)

    for reg in [1e-1, 1e-3, 1e-6]:
        B = field_inverse_tikhonov(A, reg)
        BB = np.empty_like(B)

        for i in range(grid.size):
            BB[..., i] = inverse_tikhonov(A[..., i], reg)

        assert np.allclose(B, BB)

def test_field_inverse_truncated():
    grid = make_pupil_grid(2)

    A = Field(np.random.randn(3, 3, grid.size), grid)

    for reg in [1e-1, 1e-3, 1e-6]:
        B = field_inverse_truncated(A, reg)
        BB = np.empty_like(B)

        for i in range(grid.size):
            BB[..., i] = inverse_truncated(A[..., i], reg)

        assert np.allclose(B, BB)

def test_field_inverse_truncated_modal():
    grid = make_pupil_grid(2)

    A = Field(np.random.randn(3, 3, grid.size), grid)

    for num_modes in [1, 2]:
        B = field_inverse_truncated_modal(A, num_modes)
        BB = np.empty_like(B)

        for i in range(grid.size):
            BB[..., i] = inverse_truncated_modal(np.asarray(A[..., i]), num_modes)

        assert np.allclose(B, BB)

def test_field_cross():
    grid = make_pupil_grid(2)

    A = Field(np.random.randn(3, grid.size), grid)
    B = Field(np.random.randn(3, grid.size), grid)

    C = field_cross(A, B)
    CC = np.empty_like(C)
    for i in range(grid.size):
        CC[..., i] = np.cross(A[:, i], B[:, i])

    assert np.allclose(C, CC)

def test_field_svd():
    grid = make_pupil_grid(2)

    A = Field(np.random.randn(5, 10, grid.size), grid)

    U, S, Vh = field_svd(A)
    u, s, vh = field_svd(A, False)

    for i in range(grid.size):
        svd = np.linalg.svd(A[..., i])

        assert np.allclose(U[..., i], svd[0])
        assert np.allclose(S[..., i], svd[1])
        assert np.allclose(Vh[..., i], svd[2])

        svd2 = np.linalg.svd(A[..., i], full_matrices=False)

        assert np.allclose(u[..., i], svd2[0])
        assert np.allclose(s[..., i], svd2[1])
        assert np.allclose(vh[..., i], svd2[2])

def test_grid_hashing_and_comparison():
    grid1 = make_pupil_grid(128)

    grid2 = CartesianGrid(SeparatedCoords(copy.deepcopy(grid1.separated_coords)))
    assert hash(grid1) != hash(grid2)
    assert grid1 != grid2
    assert grid2 != grid1

    grid3 = CartesianGrid(UnstructuredCoords(copy.deepcopy(grid1.coords)))
    assert hash(grid1) != hash(grid3)
    assert grid1 != grid3
    assert grid3 != grid1
    assert grid2 != grid3
    assert grid3 != grid2

    grid4 = make_pupil_grid(128)
    print('start')
    assert hash(grid1) == hash(grid4)
    assert grid1 == grid4

    grid5 = PolarGrid(grid1.coords)
    assert hash(grid1) != hash(grid5)
    assert grid1 != grid5
    assert grid5 != grid1

    grid6 = CartesianGrid(copy.deepcopy(grid1.coords))
    assert hash(grid1) == hash(grid6)
    assert grid1 == grid6

    grid7 = grid1.scaled(2)
    assert hash(grid1) != hash(grid7)
    assert grid1 != grid7

    grid8 = grid1.scaled(2)
    assert hash(grid1) != hash(grid8)
    assert hash(grid7) == hash(grid8)
    assert grid1 != grid8
    assert grid7 == grid8

    grid9 = make_pupil_grid(256)
    assert hash(grid1) != hash(grid9)
    assert grid1 != grid9

    grid10 = CartesianGrid(SeparatedCoords(copy.deepcopy(grid2.separated_coords)))
    assert hash(grid2) == hash(grid10)
    assert grid2 == grid10

    assert grid1 != 0
    assert grid1 != 'string'

def test_grid_supersampled():
    g = make_uniform_grid(128, [1, 1])
    g2 = make_supersampled_grid(g, 4)
    g3 = make_subsampled_grid(g2, 4)

    assert np.allclose(g.x, g3.x)
    assert np.allclose(g.y, g3.y)

    g4 = make_supersampled_grid(make_supersampled_grid(g, 2), 2)

    assert np.allclose(g2.x, g4.x)
    assert np.allclose(g2.y, g4.y)

@contextlib.contextmanager
def temporary_config(key, value):
    conf = Configuration()

    keys = key.split('.')

    for k in keys[:-1]:
        conf = getattr(conf, k)

    old_value = getattr(conf, keys[-1])

    setattr(conf, keys[-1], value)

    yield

    setattr(conf, keys[-1], old_value)

@pytest.mark.parametrize('use_new_style_fields', [True, False])
def test_field_arithmetic(use_new_style_fields):
    grid = make_pupil_grid(16)

    M = np.random.randn(grid.size, grid.size)

    with temporary_config('core.use_new_style_fields', use_new_style_fields):
        a = Field(np.ones(grid.size), grid)
        b = Field(np.ones(grid.size), grid)

        assert np.allclose(a, b)
        assert np.allclose(a + b, 2)
        assert np.allclose(a - b, 0)
        assert np.allclose(a * b, a)

        assert is_field(a + b)
        assert is_field(a - b)
        assert is_field(a * b)
        assert is_field(np.exp(2j * a))

        assert np.ptp(a) == a.ptp()
        assert is_field(a.conj())
        assert is_field(a.conjugate())
        assert is_field(a.clip(-1, 1))
        assert is_field(a.repeat(10))

        assert a.size == a.grid.size
        assert is_field(a.astype('bool'))
        assert np.allclose(a.sum(), np.asarray(a).sum())

        a[0] = 6
        a[1:2] = 3

        assert a[0] == 6
        assert a[1] == 3

        assert not is_field(np.asarray(a))

        assert not is_field(M.dot(a))

        assert np.allclose(a.imag, 0)

@pytest.mark.parametrize('use_new_style_fields', [True, False])
def test_field_pickle(use_new_style_fields):
    grid = make_pupil_grid(16)

    with temporary_config('core.use_new_style_fields', use_new_style_fields):
        a = Field(np.ones(grid.size), grid)

        state = pickle.dumps(a)
        b = pickle.loads(state)

        assert np.allclose(a, b)
        assert a.grid == b.grid



import numpy as np
import scipy.sparse

from ..mode_basis import ModeBasis
from .deformable_mirror import DeformableMirror

class SegmentedDeformableMirror(DeformableMirror):
    '''A segmented deformable mirror.

    This deformable mirror class can simulate devices such as those
    made by IrisAO and BMC. All segments are controlled in piston,
    tip and tilt.

    Parameters
    ----------
    segments : ModeBasis
        A mode basis with all segments.
    '''
    def __init__(self, segments):
        self.segments = segments
        self.actuators = np.zeros(len(segments) * 3)
        self.input_grid = segments.grid

    @property
    def segments(self):
        '''The segments of this deformable mirror in a ModeBasis.
        '''
        return self._segments

    @segments.setter
    def segments(self, segments):
        self._segments = segments

        tip = []
        tilt = []

        for segment in segments:
            segment_mean = segment.mean()
            norm = np.mean(segment**2) - segment_mean**2

            tip_mode = segment * segment.grid.x
            if norm != 0:
                beta = ((tip_mode * segment).mean() - segment_mean * tip_mode.mean()) / norm
                tip_mode -= beta * segment

            tip_mode = scipy.sparse.csr_matrix(tip_mode)
            tip_mode.eliminate_zeros()
            tip.append(tip_mode)

            tilt_mode = segment * segment.grid.y
            if norm != 0:
                beta = ((tilt_mode * segment).mean() - segment_mean * tilt_mode.mean()) / norm
                tilt_mode -= beta * segment

            tilt_mode = scipy.sparse.csr_matrix(tilt_mode)
            tilt_mode.eliminate_zeros()
            tilt.append(tilt_mode)

        tip = ModeBasis(tip)
        tilt = ModeBasis(tilt)

        self.influence_functions = segments + tip + tilt

    def get_segment_actuators(self, segment_id):
        '''Get the actuators for an individual segment of the DM.

        Parameters
        ----------
        segment_id : int
            The index of the segment for which to get the actuators.

        Returns
        -------
        piston : scalar
            The piston of the segment in meters.
        tip : scalar
            The tip of the segment in radians.
        tilt : scalar
            The tilt of the segment in radians.
        '''
        piston = self.actuators[segment_id]
        tip = self.actuators[segment_id + len(self._segments)]
        tilt = self.actuators[segment_id + 2 * len(self._segments)]

        return (piston, tip, tilt)

    def set_segment_actuators(self, segment_id, piston, tip, tilt):
        '''Set the actuators for an individual segment of the DM.

        Parameters
        ----------
        segment_id : int
            The index of the segment for which to get the actuators.
        piston : scalar
            The piston of the segment in meters.
        tip : scalar
            The tip of the segment in radians.
        tilt : scalar
            The tilt of the segment in radians.
        '''
        self.actuators[segment_id] = piston
        self.actuators[segment_id + len(self._segments)] = tip
        self.actuators[segment_id + 2 * len(self._segments)] = tilt



import os
import platform
import multiprocessing

def get_num_available_cores():
    '''Get the number of cores available on the system.

    The attempt at retrieving the number of cores that our process
    is assigned may not be available on all operating systems. On
    unsupported operating systems, the total number of cores is
    returned instead.

    Returns
    -------
    int
        The number of available cores.

    Raises
    ------
    RuntimeError
        If we are unable to
    '''
    if hasattr(os, 'sched_getaffinity'):
        return len(os.sched_getaffinity(0))
    elif platform.system() == 'Windows':
        import ctypes
        import ctypes.wintypes

        kernel32 = ctypes.WinDLL('kernel32')

        DWORD_PTR = ctypes.wintypes.WPARAM
        PDWORD_PTR = ctypes.POINTER(DWORD_PTR)

        GetCurrentProcess = kernel32.GetCurrentProcess
        GetCurrentProcess.restype = ctypes.wintypes.HANDLE

        GetProcessAffinityMask = kernel32.GetProcessAffinityMask
        GetProcessAffinityMask.argtypes = (ctypes.wintypes.HANDLE, PDWORD_PTR, PDWORD_PTR)

        mask = DWORD_PTR()

        if GetProcessAffinityMask(GetCurrentProcess(), ctypes.byref(mask), ctypes.byref(DWORD_PTR())):
            # Call successful. Return result.
            return bin(mask.value).count('1')

    # Operating system is unsupported or the call to retrieve the
    # core count has failed. Fall back to all cores.
    return multiprocessing.cpu_count()



__all__ = [
    'Propagator',
    'FresnelPropagator',
    'FraunhoferPropagator',
    'AngularSpectrumPropagator'
]

from .propagator import *
from .angular_spectrum import *
from .fresnel import *
from .fraunhofer import *



import numpy as np
import os
from hcipy import *

def test_grid_io():
    grid1 = make_pupil_grid(128)
    grid2 = CartesianGrid(SeparatedCoords(grid1.separated_coords))
    grid3 = grid1.as_('polar')

    grids = [grid1, grid2, grid3]
    formats = ['asdf', 'fits', 'fits.gz', 'pkl', 'pickle']
    filenames = ['grid_test.' + fmt for fmt in formats]

    for g in grids:
        for fname in filenames:
            write_grid(g, fname)
            new_grid = read_grid(fname)

            assert hash(g) == hash(new_grid)

            os.remove(fname)

def test_field_io():
    grid = make_pupil_grid(128)
    field = make_circular_aperture(1)(grid)

    formats = ['asdf', 'fits', 'fits.gz', 'pkl', 'pickle']
    filenames = ['field_test.' + fmt for fmt in formats]

    for fname in filenames:
        write_field(field, fname)
        new_field = read_field(fname)

        assert np.allclose(field, new_field)
        assert hash(field.grid) == hash(new_field.grid)

        os.remove(fname)

def test_mode_basis_io():
    grid = make_pupil_grid(128)
    mode_bases = [
        make_zernike_basis(20, 1, grid, 1),
        make_xinetics_influence_functions(grid, 8, 1 / 8)
    ]

    formats = ['asdf', 'fits', 'fits.gz', 'pkl', 'pickle']
    filenames = ['mode_basis_test.' + fmt for fmt in formats]

    for mode_basis in mode_bases:
        for fname in filenames:
            write_mode_basis(mode_basis, fname)
            new_mode_basis = read_mode_basis(fname)

            assert hash(new_mode_basis.grid) == hash(mode_basis.grid)
            assert new_mode_basis.is_sparse == mode_basis.is_sparse

            for i in range(mode_basis.num_modes):
                assert np.allclose(mode_basis[i], new_mode_basis[i])

            os.remove(fname)


def test_finite_difference():

    grid = make_pupil_grid(16)
    surface = grid.ones()

    Dx = make_derivative_matrix(grid, axis='x')
    Dy = make_derivative_matrix(grid, axis='y')

    surface_dx = Dx.dot(surface)
    surface_dy = Dy.dot(surface)

    assert abs(np.median(surface_dx)) < 1e-10
    assert abs(np.median(surface_dy)) < 1e-10

    laplacian = make_laplacian_matrix(grid)
    surface_lap = laplacian.dot(surface)
    assert abs(np.median(surface_lap)) < 1e-10

    # Test if numpy array is handled correctly b
    laplacian_kernel = np.array([[0, 1, 0], [1, -4, 1], [0, 1, 0]])
    laplacian_operator = generate_convolution_matrix(grid, laplacian_kernel)
    surface_lap = laplacian_operator.dot(surface)
    assert abs(np.median(surface_lap)) < 1e-10

def test_poisson():
    num_trials = 100
    num_runs = 100000
    lam = 100.0
    sigma = np.sqrt(lam / num_trials)

    lam_realization = large_poisson(lam * np.ones((num_trials, num_runs)), thresh=1e6)

    assert (np.std(np.mean(lam_realization, axis=0) - lam) - sigma) / sigma < 1e-2

    num_trials = 100
    num_runs = 100000
    lam = 10000000.0
    sigma = np.sqrt(lam / num_trials)

    lam_realization = large_poisson(lam * np.ones((num_trials, num_runs)), thresh=1e6)

    assert (np.std(np.mean(lam_realization, axis=0) - lam) - sigma) / sigma < 1e-2

def test_gamma():
    num_trials = 100
    num_runs = 100000
    lam = 100.0
    theta = 1 / 10.0
    mean = lam * theta
    sigma = np.sqrt(lam * theta**2 / num_trials)

    lam_realization = large_gamma(lam * np.ones((num_trials, num_runs)), theta, thresh=1e6)

    assert (np.std(np.mean(lam_realization, axis=0) - mean) - sigma) / sigma < 1e-2

    num_trials = 100
    num_runs = 100000
    lam = 10000000.0
    theta = 1 / 10.0

    mean = lam * theta
    sigma = np.sqrt(lam * theta**2 / num_trials)

    lam_realization = large_gamma(lam * np.ones((num_trials, num_runs)), theta, thresh=1e6)

    assert (np.std(np.mean(lam_realization, axis=0) - mean) - sigma) / sigma < 1e-2

def test_emccd_noise():
    photo_electron_flux = 1000.0
    read_noise = 0
    emgain = 500

    num_trials = 100
    num_runs = 100000

    sigma = np.sqrt(2 * photo_electron_flux / num_trials)
    noise = make_emccd_noise(photo_electron_flux * np.ones((num_trials, num_runs)), read_noise, emgain)

    assert abs(np.std(np.mean(noise, axis=0) / emgain - photo_electron_flux) - sigma) / sigma < 1e-2



import numpy as np
import matplotlib.pyplot as plt
from hcipy import *
import pytest
import os

def test_agnostic_apodizer():
    aperture_achromatic = make_circular_aperture(1)
    aperture_chromatic = lambda input_grid, wavelength: make_circular_aperture(wavelength)(input_grid)

    apod_achromatic = Apodizer(aperture_achromatic)
    apod_chromatic = Apodizer(aperture_chromatic)

    phase_chromatic = lambda input_grid, wavelength: zernike(2, 0)(input_grid) * wavelength
    apod_phase_chromatic = PhaseApodizer(phase_chromatic)

    phase_achromatic = zernike(2, 0)
    apod_phase_achromatic = PhaseApodizer(phase_achromatic)

    filter_curve = lambda wavelength: np.sqrt(wavelength)
    apod_filter = Apodizer(filter_curve)

    for wl in [0.2, 0.4, 0.7]:
        for num_pix in [32, 64, 128]:
            for diameter in [0.5, 1, 2]:
                pupil_grid = make_pupil_grid(num_pix, diameter)
                wf = Wavefront(pupil_grid.ones(), wl)

                pup = apod_achromatic(wf).electric_field
                assert np.allclose(pup, aperture_achromatic(pupil_grid))

                pup = apod_chromatic(wf).electric_field
                assert np.allclose(pup, aperture_chromatic(pupil_grid, wl))

                pup = apod_phase_achromatic(wf).electric_field
                assert np.allclose(pup, np.exp(1j * phase_achromatic(pupil_grid)))

                pup = apod_phase_chromatic(wf).electric_field
                assert np.allclose(pup, np.exp(1j * phase_chromatic(pupil_grid, wl)))

                pup = apod_filter(wf).electric_field
                assert np.allclose(pup, filter_curve(wl))

def test_statistics_noisy_detector():
    N = 256
    grid = make_pupil_grid(N)

    field = Field(np.ones(N**2), grid)

    # First we test photon noise, dark current noise and read noise.
    flat_field = 0
    dark_currents = np.logspace(1, 6, 6)
    read_noises = np.logspace(1, 6, 6)
    photon_noise = True

    for dc in dark_currents:
        for rn in read_noises:
            # The test detector.
            detector = NoisyDetector(detector_grid=grid, include_photon_noise=photon_noise, flat_field=flat_field, dark_current_rate=dc, read_noise=rn)

            # The integration times we will test.
            integration_time = np.logspace(1, 6, 6)

            for t in integration_time:
                # integration
                detector.integrate(field, t)

                # read out
                measurement = detector.read_out()

                # The std of the data by the detector.
                std_measurement = np.std(measurement)

                # The std that we expect given the input.
                expected_std = np.sqrt(field[0] * t + rn**2 + dc * t)

                assert np.isclose(expected_std, std_measurement, rtol=2e-02, atol=1e-05)

    # Now we test the flat field separately.
    flat_fields = np.linspace(0, 1, 100)
    dark_current = 0
    read_noise = 0
    photon_noise = False

    for ff in flat_fields:
        detector = NoisyDetector(detector_grid=grid, include_photon_noise=photon_noise, flat_field=ff, dark_current_rate=dark_current, read_noise=read_noise)

        integration_times = np.logspace(1, 6, 6)

        for t in integration_times:
            # integration and read out
            detector.integrate(field, t)
            measurement = detector.read_out()

            # The std of the data by the detector.
            std_measurement = np.std(measurement)

            # The std that we expect given the input.
            expected_std = ff * field[0] * t

            assert np.isclose(expected_std, std_measurement, rtol=2e-02, atol=1e-05)

    aperture = make_circular_aperture(1)(grid)
    subsamplings = [1, 2, 4, 8]

    for subsampling in subsamplings:
        grid_subsampled = make_subsampled_grid(grid, subsampling)
        aperture_subsampled = subsample_field(aperture, subsampling, grid_subsampled, statistic='sum')

        # the detector with the new subsampling factor
        detector = NoisyDetector(detector_grid=grid_subsampled, include_photon_noise=False, subsampling=subsampling)

        # integrating and reading out
        detector.integrate(aperture, dt=1)
        detector_image = detector.read_out()

        # testing if the image from the detector matches the subsampled aperture
        assert np.allclose(aperture_subsampled, detector_image)

def test_glass_catalogue():
    bk7 = get_refractive_index('N-BK7')
    assert np.allclose(bk7(500e-9), 1.5214144761028994)

    with pytest.raises(ValueError) as exception_info:
        get_refractive_index('N-Bk7')
    assert 'Did you mean' in str(exception_info.value)

    with pytest.raises(ValueError) as exception_info:
        get_refractive_index('N-Ba7')
    assert 'Did you mean' not in str(exception_info.value)

def test_deformable_mirror():
    num_pix = 256
    x_tilt = np.radians(15)
    y_tilt = np.radians(10)
    z_tilt = np.radians(20)
    grid = make_pupil_grid(num_pix, 1.5)

    functions = [make_gaussian_influence_functions, make_xinetics_influence_functions]

    for num_actuators_across_pupil in [12, 16]:
        actuator_spacing = 1 / num_actuators_across_pupil
        num_actuators = num_actuators_across_pupil**2

        # Check if the actuator spacing of an unrotated DM is correct
        actuator_positions = make_actuator_positions(num_actuators_across_pupil, actuator_spacing)
        assert np.allclose(actuator_positions.delta, actuator_spacing)

        actuator_positions = make_actuator_positions(num_actuators_across_pupil, actuator_spacing, x_tilt=x_tilt, y_tilt=y_tilt, z_tilt=z_tilt)

        for func in functions:
            influence_functions = func(grid, num_actuators_across_pupil, actuator_spacing, x_tilt=x_tilt, y_tilt=y_tilt, z_tilt=z_tilt)

            deformable_mirror = DeformableMirror(influence_functions)

            # Check number of generated influence functions
            assert len(influence_functions) == num_actuators
            assert deformable_mirror.num_actuators == num_actuators

            # Check that influence functions are sparse
            assert influence_functions.is_sparse

            # Check if the centroids for each influence function are close to the predicted position.
            for act, p in zip(influence_functions, actuator_positions.points):
                x_pos = (act * grid.x).sum() / act.sum()
                y_pos = (act * grid.y).sum() / act.sum()

                assert np.allclose([x_pos, y_pos], p, atol=0.02)

            # Mirror should start out flattened
            assert np.std(deformable_mirror.surface) < 1e-12

            for i in np.random.randint(0, num_actuators, size=10):
                deformable_mirror.actuators[i] = np.random.randn(1)[0]

                # Mirror should not be flat with poked actuator
                assert np.std(deformable_mirror.surface) > 1e-12

                deformable_mirror.actuators[i] = 0

            # Mirror should be flat again
            assert np.std(deformable_mirror.surface) < 1e-12

            deformable_mirror.random(0.001)

            # Mirror should have random actuator pokes
            assert np.std(deformable_mirror.surface) > 1e-12

            wf = Wavefront(grid.ones())
            wf_out = deformable_mirror.forward(wf)

            # Check that the correct phase is applied
            assert np.allclose(wf_out.phase, deformable_mirror.phase_for(1))

            wf_in = deformable_mirror.backward(wf_out)

            # Check that the deformable mirror is phase only
            assert np.allclose(wf_in.electric_field, wf.electric_field)

            # Check OPD
            assert np.allclose(deformable_mirror.opd, 2 * deformable_mirror.surface)

            deformable_mirror.flatten()

            # Check that the deformable mirror is flat again
            assert np.std(deformable_mirror.surface) < 1e-12

def test_segmented_deformable_mirror():
    num_pix = 256
    grid = make_pupil_grid(num_pix)

    for num_rings in [2, 4, 5]:
        num_segments_expected = 3 * (num_rings + 1) * num_rings + 1

        segment_positions = make_hexagonal_grid(0.5 / num_rings * np.sqrt(3) / 2, num_rings)
        aperture, segments = make_segmented_aperture(make_hexagonal_aperture(0.5 / num_rings - 0.003, np.pi / 2), segment_positions, return_segments=True)

        aperture = evaluate_supersampled(aperture, grid, 2)
        segments = evaluate_supersampled(segments, grid, 2)

        # Check number of generated segments
        assert len(segments) == num_segments_expected

        segmented_mirror = SegmentedDeformableMirror(segments)

        # Mirror should start out flattened.
        assert np.std(segmented_mirror.surface) < 1e-12

        for i in np.random.randint(0, num_segments_expected, size=10):
            piston = np.random.randn(1)[0]

            while np.abs(piston) < 1e-5:
                piston = np.random.randn(1)[0]

            segmented_mirror.set_segment_actuators(i, piston, 0, 0)

            # Mirror should have the correct piston
            assert np.abs(segmented_mirror.get_segment_actuators(i)[0] - piston) / piston < 1e-5
            assert np.abs(np.ptp(segmented_mirror.surface) - piston) / piston < 1e-5

            tip, tilt = np.random.randn(2)

            while np.abs(tip) < 1e-5:
                tip = np.random.randn(1)[0]

            while np.abs(tilt) < 1e-5:
                tilt = np.random.randn(1)[0]

            segmented_mirror.set_segment_actuators(i, 0, tip, tilt)

            # Mirror should be distorted with tip and tilt on a segment
            assert np.std(segmented_mirror.surface > 1e-12)

            # Return segment to zero position
            segmented_mirror.set_segment_actuators(i, 0, 0, 0)

        # Run labeling of segments
        imshow_field(aperture)
        label_actuator_centroid_positions(segments)
        plt.clf()

def test_wavefront_stokes():
    N = 4
    grid = make_pupil_grid(N)

    stokes = np.random.uniform(-1, 1, 4)
    stokes[0] = np.abs(stokes[0]) + np.linalg.norm(stokes[1:])

    amplitude = np.random.uniform(0, 1, (2, 2, N**2))
    phase = np.random.uniform(0, 2 * np.pi, (2, 2, N**2))

    jones_field = Field(amplitude * np.exp(1j * phase), grid)

    determinants = field_determinant(jones_field)
    jones_field[:, :, determinants > 1] *= np.random.uniform(0, 1, np.sum(determinants > 1)) / determinants[determinants > 1]

    jones_element = JonesMatrixOpticalElement(jones_field)

    mueller_field = jones_to_mueller(jones_field)

    field = Field(np.ones(N**2), grid)
    stokes_wavefront = Wavefront(field, input_stokes_vector=stokes)

    jones_element_forward = jones_element.forward(stokes_wavefront)
    mueller_forward = field_dot(mueller_field, stokes)

    assert np.allclose(jones_element_forward.I, mueller_forward[0])
    assert np.allclose(jones_element_forward.Q, mueller_forward[1])
    assert np.allclose(jones_element_forward.U, mueller_forward[2])
    assert np.allclose(jones_element_forward.V, mueller_forward[3])

def test_degree_and_angle_of_polarization():
    grid = make_pupil_grid(16)

    wf = Wavefront(grid.ones(), input_stokes_vector=[1, 0, 0, 0])
    assert np.allclose(wf.degree_of_polarization, 0)
    assert np.allclose(wf.degree_of_linear_polarization, 0)
    assert np.allclose(wf.degree_of_circular_polarization, 0)

    wf = Wavefront(grid.ones(), input_stokes_vector=[1, 1, 0, 0])
    assert np.allclose(wf.degree_of_polarization, 1)
    assert np.allclose(wf.degree_of_linear_polarization, 1)
    assert np.allclose(wf.angle_of_linear_polarization, 0)
    assert np.allclose(wf.degree_of_circular_polarization, 0)

    wf = Wavefront(grid.ones(), input_stokes_vector=[1, 0.5, 0, 0])
    assert np.allclose(wf.degree_of_polarization, 0.5)
    assert np.allclose(wf.degree_of_linear_polarization, 0.5)
    assert np.allclose(wf.angle_of_linear_polarization, 0)
    assert np.allclose(wf.degree_of_circular_polarization, 0)

    wf = Wavefront(grid.ones(), input_stokes_vector=[1, -1, 0, 0])
    assert np.allclose(wf.degree_of_polarization, 1)
    assert np.allclose(wf.degree_of_linear_polarization, 1)
    assert np.allclose(wf.angle_of_linear_polarization, np.pi / 2)
    assert np.allclose(wf.degree_of_circular_polarization, 0)

    wf = Wavefront(grid.ones(), input_stokes_vector=[1, 0, 1, 0])
    assert np.allclose(wf.degree_of_polarization, 1)
    assert np.allclose(wf.degree_of_linear_polarization, 1)
    assert np.allclose(wf.angle_of_linear_polarization, np.pi / 4)
    assert np.allclose(wf.degree_of_circular_polarization, 0)

    wf = Wavefront(grid.ones(), input_stokes_vector=[1, 0, 0, 1])
    assert np.allclose(wf.degree_of_polarization, 1)
    assert np.allclose(wf.degree_of_linear_polarization, 0)
    assert np.allclose(wf.degree_of_circular_polarization, 1)

    wf = Wavefront(grid.ones(), input_stokes_vector=[1, 0, 0, 0.5])
    assert np.allclose(wf.degree_of_polarization, 0.5)
    assert np.allclose(wf.degree_of_linear_polarization, 0)
    assert np.allclose(wf.degree_of_circular_polarization, 0.5)

    wf = Wavefront(grid.ones(), input_stokes_vector=[1, 0, 0, -1])
    assert np.allclose(wf.degree_of_polarization, 1)
    assert np.allclose(wf.degree_of_linear_polarization, 0)
    assert np.allclose(wf.degree_of_circular_polarization, -1)

    wf = Wavefront(grid.ones(), input_stokes_vector=[2, np.sqrt(2), 0, np.sqrt(2)])
    assert np.allclose(wf.degree_of_polarization, 1)
    assert np.allclose(wf.degree_of_linear_polarization, 1 / np.sqrt(2))
    assert np.allclose(wf.degree_of_circular_polarization, 1 / np.sqrt(2))

def mueller_matrix_for_general_linear_retarder(theta, delta):
    '''Analytic expression Mueller matrix linear retarder.

    Parameters
    ----------
    theta : scaler
        rotation angle optic in radians
    delta : scalar
        retardance optic in radians
    '''
    retarder = np.zeros((4, 4))

    retarder[0, 0] = 1

    retarder[1, 1] = np.cos(2 * theta)**2 + np.sin(2 * theta)**2 * np.cos(delta)
    retarder[1, 2] = np.cos(2 * theta) * np.sin(2 * theta) * (1 - np.cos(delta))
    retarder[1, 3] = np.sin(2 * theta) * np.sin(delta)

    retarder[2, 1] = np.cos(2 * theta) * np.sin(2 * theta) * (1 - np.cos(delta))
    retarder[2, 2] = np.cos(2 * theta)**2 * np.cos(delta) + np.sin(2 * theta)**2
    retarder[2, 3] = -np.cos(2 * theta) * np.sin(delta)

    retarder[3, 1] = -np.sin(2 * theta) * np.sin(delta)
    retarder[3, 2] = np.cos(2 * theta) * np.sin(delta)
    retarder[3, 3] = np.cos(delta)

    return retarder

def mueller_matrix_for_general_linear_polarizer(theta):
    '''Analytic expression Mueller matrix linear polarizer.

    Parameters
    ----------
    theta : scaler
        rotation angle optic in radians
    '''
    polarizer = np.zeros((4, 4))

    polarizer[0, 0] = 1
    polarizer[0, 1] = np.cos(2 * theta)
    polarizer[0, 2] = np.sin(2 * theta)

    polarizer[1, 0] = np.cos(2 * theta)
    polarizer[1, 1] = np.cos(2 * theta)**2
    polarizer[1, 2] = 0.5 * np.sin(4 * theta)

    polarizer[2, 0] = np.sin(2 * theta)
    polarizer[2, 1] = 0.5 * np.sin(4 * theta)
    polarizer[2, 2] = np.sin(2 * theta)**2

    polarizer *= 0.5
    return polarizer

def test_polarization_elements():
    N = 1
    grid = make_pupil_grid(N)
    test_field = grid.ones()

    # Stokes vectors used for testing.
    stokes_vectors = [
        None,
        np.array([1, 0, 0, 0]),  # unpolarized
        np.array([1, 1, 0, 0]),  # +Q polarized
        np.array([1, -1, 0, 0]),  # -Q polarized
        np.array([1, 0, 1, 0]),  # +U polarized
        np.array([1, 0, -1, 0]),  # -U polarized
        np.array([1, 0, 0, 1]),  # +V polarized
        np.array([1, 0, 0, -1])  # -V polarized
    ]

    # Angles of the optics that will be tested.
    angles = [-45, -22.5, 0, 22.5, 45, 90]  # degrees

    # Test QWPs, HWPs, polarizers and PBSs
    for stokes_vector in stokes_vectors:
        test_wf = Wavefront(test_field, input_stokes_vector=stokes_vector)

        if stokes_vector is None:
            # Set Stokes vector for futher calculation to unpolarized light.
            stokes_vector = np.array([1, 0, 0, 0])

        for angle in angles:
            # Create quarterwave plate.
            QWP_hcipy = QuarterWavePlate(np.radians(angle))

            # Test if Mueller matrix is the same as reference.
            QWP_ref = mueller_matrix_for_general_linear_retarder(np.radians(angle), -np.pi / 2)
            assert np.allclose(QWP_hcipy.mueller_matrix, QWP_ref)

            # Propagate wavefront through optical element.
            wf_forward_QWP = QWP_hcipy.forward(test_wf)

            # Test if result is the same as reference.
            reference_stokes_post_QWP = field_dot(QWP_ref, stokes_vector)
            assert np.allclose(wf_forward_QWP.stokes_vector[:, 0], reference_stokes_post_QWP)

            # Test backward propagation.
            wf_forward_backward_QWP = QWP_hcipy.backward(wf_forward_QWP)
            assert np.allclose(wf_forward_backward_QWP.stokes_vector[:, 0], stokes_vector)

            # Test power conservation.
            assert np.allclose(wf_forward_QWP.I, test_wf.I)

            # Create halfwave plate
            HWP_hcipy = HalfWavePlate(np.radians(angle))

            # Test if Mueller matrix is the same as reference.
            HWP_ref = mueller_matrix_for_general_linear_retarder(np.radians(angle), -np.pi)
            assert np.allclose(HWP_hcipy.mueller_matrix, HWP_ref)

            # Propagate wavefront through optical element.
            wf_forward_HWP = HWP_hcipy.forward(test_wf)

            # Test if result is the same as reference.
            reference_stokes_post_HWP = field_dot(HWP_ref, stokes_vector)
            assert np.allclose(wf_forward_HWP.stokes_vector[:, 0], reference_stokes_post_HWP)

            # Test backward propagation.
            wf_forward_backward_HWP = HWP_hcipy.backward(wf_forward_HWP)
            assert np.allclose(wf_forward_backward_HWP.stokes_vector[:, 0], stokes_vector)

            # Test power conservation.
            assert np.allclose(wf_forward_HWP.I, test_wf.I)

            # Create polarizer.
            polarizer_hcipy = LinearPolarizer(np.radians(angle))

            # Test if Mueller matrix is the same as reference.
            polarizer_ref = mueller_matrix_for_general_linear_polarizer(np.radians(angle))
            assert np.allclose(polarizer_hcipy.mueller_matrix, polarizer_ref)

            # Propagate wavefront through optical element.
            wf_forward_polarizer = polarizer_hcipy.forward(test_wf)

            # Test if result is the same as reference.
            reference_stokes_post_polarizer = field_dot(polarizer_ref, stokes_vector)
            assert np.allclose(wf_forward_polarizer.stokes_vector[:, 0], reference_stokes_post_polarizer)

            # Test backward propagation.
            wf_backward_polarizer = polarizer_hcipy.backward(test_wf)
            assert np.allclose(wf_backward_polarizer.stokes_vector[:, 0], reference_stokes_post_polarizer)

            # Create polarizing beam splitter
            LPBS_hcipy = LinearPolarizingBeamSplitter(np.radians(angle))

            # Test if Mueller matrices are the same as reference.
            polarizer_1_ref = mueller_matrix_for_general_linear_polarizer(np.radians(angle))
            polarizer_2_ref = mueller_matrix_for_general_linear_polarizer(np.radians(angle + 90))
            assert np.allclose(LPBS_hcipy.mueller_matrices[0], polarizer_1_ref)
            assert np.allclose(LPBS_hcipy.mueller_matrices[1], polarizer_2_ref)

            # Propagate wavefront through optical element.
            wf_forward_polarizer_1, wf_forward_polarizer_2 = LPBS_hcipy.forward(test_wf)

            # Test if result is the same as reference.
            reference_stokes_post_polarizer_1 = field_dot(polarizer_1_ref, stokes_vector)
            reference_stokes_post_polarizer_2 = field_dot(polarizer_2_ref, stokes_vector)
            assert np.allclose(wf_forward_polarizer_1.stokes_vector[:, 0], reference_stokes_post_polarizer_1)
            assert np.allclose(wf_forward_polarizer_2.stokes_vector[:, 0], reference_stokes_post_polarizer_2)

            # Test power conservation.
            assert np.allclose(test_wf.I, wf_forward_polarizer_1.I + wf_forward_polarizer_2.I)

            # Test multiplication of polarization optics
            # 1) JonesMatrixOpticalElement * JonesMatrixOpticalElement
            multiplication_test_1 = polarizer_hcipy * QWP_hcipy
            # 2) JonesMatrixOpticalElement * numpy array
            multiplication_test_2 = polarizer_hcipy * QWP_hcipy.jones_matrix

            multiplication_test_ref = np.dot(polarizer_ref, QWP_ref)

            # testing if the Mueller matrices are the same
            assert np.allclose(multiplication_test_1.mueller_matrix, multiplication_test_ref)
            assert np.allclose(multiplication_test_2.mueller_matrix, multiplication_test_ref)

            # propagating the wavefront through the optics
            wf_forward_multiplication_1 = multiplication_test_1.forward(test_wf)
            wf_forward_multiplication_2 = multiplication_test_2.forward(test_wf)

            reference_stokes_post_multiplication = field_dot(multiplication_test_ref, stokes_vector)

            assert np.allclose(wf_forward_multiplication_1.stokes_vector[:, 0], reference_stokes_post_multiplication)
            assert np.allclose(wf_forward_multiplication_2.stokes_vector[:, 0], reference_stokes_post_multiplication)

        # Create polarizing beam splitter
        CPBS_hcipy = CircularPolarizingBeamSplitter()

        # Test if Mueller matrices are the same as reference.
        circ_polarizer_1_ref = mueller_matrix_for_general_linear_polarizer(0)
        circ_polarizer_2_ref = mueller_matrix_for_general_linear_polarizer(np.radians(90))
        QWP_1_ref = mueller_matrix_for_general_linear_retarder(np.pi / 4, -np.pi / 2)
        CP_1_ref = np.dot(circ_polarizer_1_ref, QWP_1_ref)
        CP_2_ref = np.dot(circ_polarizer_2_ref, QWP_1_ref)

        assert np.allclose(CPBS_hcipy.mueller_matrices[0], CP_1_ref)
        assert np.allclose(CPBS_hcipy.mueller_matrices[1], CP_2_ref)

        # Propagate wavefront through optical element.
        wf_forward_circ_polarizer_1, wf_forward_circ_polarizer_2 = CPBS_hcipy.forward(test_wf)

        # Test if result is the same as reference.
        reference_stokes_post_circ_polarizer_1 = field_dot(CP_1_ref, stokes_vector)
        reference_stokes_post_circ_polarizer_2 = field_dot(CP_2_ref, stokes_vector)
        assert np.allclose(wf_forward_circ_polarizer_1.stokes_vector[:, 0], reference_stokes_post_circ_polarizer_1)
        assert np.allclose(wf_forward_circ_polarizer_2.stokes_vector[:, 0], reference_stokes_post_circ_polarizer_2)

        # Test power conservation.
        assert np.allclose(test_wf.I, wf_forward_circ_polarizer_1.I + wf_forward_circ_polarizer_2.I)

def test_magnifier():
    pupil_grid = make_pupil_grid(128)
    wf = Wavefront(make_circular_aperture(1)(pupil_grid))
    wf.total_power = 1

    magnifier = Magnifier(1.0)

    for magnification in [0.2, 3.0, [0.3, 0.3], [0.5, 2]]:
        magnifier.magnification = magnification

        wf_backward = magnifier.backward(wf)
        assert np.abs(wf_backward.total_power - 1) < 1e-12
        assert hash(wf_backward.electric_field.grid) == hash(magnifier.get_input_grid(wf.electric_field.grid, 1))

        wf_forward = magnifier.forward(wf)
        assert np.abs(wf_forward.total_power - 1) < 1e-12
        assert hash(wf_forward.electric_field.grid) == hash(magnifier.get_output_grid(wf.electric_field.grid, 1))

@pytest.mark.xfail(reason='known difficult bug; fix in progress')
def test_pickle_optical_element():
    import dill as pickle

    fname = 'optical_element.pkl'

    pupil_grid = make_pupil_grid(512)
    focal_grid = make_focal_grid(4, 16)

    elem1 = FraunhoferPropagator(pupil_grid, focal_grid)
    elem2 = SurfaceApodizer(pupil_grid.ones(), lambda wvl: 1.5 + wvl)
    elems = [elem1, elem2]

    for elem in elems:
        wf = Wavefront(pupil_grid.zeros())
        img_saved = elem(wf)

        with open(fname, 'wb') as f:
            pickle.dump(elem, f)

        with open(fname, 'rb') as f:
            elem_loaded = pickle.load(f)
            img_loaded = elem_loaded(wf)

            assert np.allclose(img_saved.electric_field, img_loaded.electric_field)

        os.remove(fname)

def test_step_index_fiber():
    core_radius_multimode = 25e-6  # m
    core_radius_singlemode = 2e-6  # m
    fiber_na = 0.13
    fiber_length = 10  # m
    D_pupil = 1  # m
    wavelength = 1e-6  # m

    multimode_fiber = StepIndexFiber(core_radius_multimode, fiber_na, fiber_length)
    singlemode_fiber = StepIndexFiber(core_radius_singlemode, fiber_na, fiber_length)

    assert np.allclose(multimode_fiber.num_modes(wavelength), 208.4953929730128)
    assert np.allclose(singlemode_fiber.num_modes(wavelength), 1.3343705150272818)

    assert np.allclose(multimode_fiber.V(wavelength), 20.42035224833366)
    assert np.allclose(singlemode_fiber.V(wavelength), 1.6336281798666927)

    assert np.allclose(multimode_fiber.mode_field_radius(wavelength), 1.6688624611223803e-05)
    assert np.allclose(singlemode_fiber.mode_field_radius(wavelength), 3.153705704872073e-06)

    core_radii = [core_radius_singlemode, core_radius_multimode]
    fibers = [singlemode_fiber, multimode_fiber]

    for fiber, core_radius in zip(fibers, core_radii):
        pupil_grid = make_pupil_grid(128)
        focal_grid = make_pupil_grid(128, 4 * core_radius)
        focal_length = D_pupil / (2 * fiber_na)

        prop = FraunhoferPropagator(pupil_grid, focal_grid, focal_length=focal_length)

        wf = Wavefront(make_circular_aperture(D_pupil)(pupil_grid), wavelength)
        wf.total_power = 1
        img = prop(wf)

        forward_wf = fiber.forward(img)
        backward_wf = fiber.backward(img)

        assert forward_wf.total_power <= img.total_power * (1 + 1e-4)
        assert backward_wf.total_power <= forward_wf.total_power * (1 + 1e-4)

def test_gaussian_fiber_mode_power():
    for q in [16, 32, 64]:
        for mfd in [0.25, 0.5, 1]:
            grid = make_focal_grid(q, 4)
            mode = make_gaussian_fiber_mode(mfd)(grid)

            wf = Wavefront(mode)

            assert np.allclose(wf.total_power, 1)

def test_gaussian_fiber_mode_mfd():
    for mfd in [0.25, 0.5, 1]:
        grid = CartesianGrid(UnstructuredCoords(([0, mfd / 2], [0, 0])))
        mode = make_gaussian_fiber_mode(mfd)(grid)

        assert np.allclose(mode[1] / mode[0], np.exp(-1))

def test_single_mode_fiber_injection():
    grid = make_focal_grid(16, 4)
    mode_1 = make_gaussian_fiber_mode(1)

    def mode_2(grid):
        return mode_1(grid) * 1j

    for mode in [mode_1, mode_2]:
        smf = SingleModeFiberInjection(grid, mode)

        wf = Wavefront(mode(grid))
        wf.total_power = 1

        post_fiber_wf = smf.forward(wf)

        # This wavefront should fully couple into the fiber.
        assert np.allclose(post_fiber_wf.electric_field, 1)
        assert np.allclose(post_fiber_wf.total_power, 1)

        pre_fiber_wf = smf.backward(post_fiber_wf)

        # Power should be conserved when injecting a perfect mode.
        assert np.allclose(pre_fiber_wf.electric_field, wf.electric_field)
        assert np.allclose(pre_fiber_wf.total_power, 1)

        wf = Wavefront(mode(grid) * grid.x)
        wf.total_power = 1

        post_fiber_wf = smf.forward(wf)

        # This wavefront should not couple at all into the fiber.
        assert np.allclose(post_fiber_wf.total_power, 0)

def check_thin_lens(wf, lens, num_steps, focal_length):
    dz = focal_length / (num_steps / 2)
    prop = FresnelPropagator(wf.electric_field.grid, dz)

    wavefront = lens(wf)

    z = [0]
    peak = [wavefront.power.max()]

    for i in range(num_steps):
        wavefront = prop(wavefront)

        z.append(z[-1] + dz)
        peak.append(wavefront.power.max())

    print('z', z)
    print('peak', peak)

    return z, peak

def test_thin_lens():
    wavelength = 1e-6
    pupil_diameter = 5e-2
    num_steps = 20

    grid = make_pupil_grid(256, 2 * pupil_diameter)

    focal_length = 300e-1
    lens = ThinLens(focal_length, lambda x: 1.5, 1e-6)

    aperture = evaluate_supersampled(make_circular_aperture(pupil_diameter), grid, 8)
    assert abs((lens.focal_length - focal_length) / focal_length) < 1e-10

    wf = Wavefront(aperture, wavelength)
    wf.total_power = 1

    z1, peak1 = check_thin_lens(wf, lens, num_steps, focal_length)

    lens.focal_length = 2 * focal_length
    z2, peak2 = check_thin_lens(wf, lens, num_steps, 2 * focal_length)

    lens.focal_length = 0.75 * focal_length
    z3, peak3 = check_thin_lens(wf, lens, num_steps, 0.75 * focal_length)

    assert abs(z1[np.argmax(peak1)] / focal_length - 1.0) < 0.01
    assert abs(z2[np.argmax(peak2)] / (2 * focal_length) - 1.0) < 0.01
    assert abs(z3[np.argmax(peak3)] / (0.75 * focal_length) - 1.0) < 0.01

def test_phase_grating():
    from scipy.special import jv
    grid = make_pupil_grid(128, 1)
    aperture = make_circular_aperture(1)(grid)

    cor = PerfectCoronagraph(aperture)
    wf = Wavefront(aperture)
    wf.total_power = 1.0

    amplitudes = np.linspace(0, np.pi, 31)
    diffraction_efficiency = np.array([cor(PhaseGrating(1 / 20, amp)(wf)).total_power for amp in amplitudes])
    eta = 1 - jv(0, amplitudes)**2
    err = eta - diffraction_efficiency

    assert np.allclose(err, 0, atol=1e-3)

    grating = PhaseGrating(1 / 20, np.pi / 2, lambda grid: np.sign(np.sin(2 * np.pi * grid.y)), orientation=0)
    wf_grating = cor(grating(wf))
    assert abs(wf_grating.total_power - 1) < 1e-15

    grating.orientation = np.pi / 2
    wf_grating = cor(grating(wf))
    assert abs(wf_grating.total_power - 1) < 1e-15

def test_thin_prism():
    nbk7 = get_refractive_index("N-BK7")
    wedge_angle = np.deg2rad(3. + 53.0 / 60.0)
    thin_prism = ThinPrism(wedge_angle, nbk7)

    test_wavelengths = np.linspace(500, 1000, 31) * 1e-9
    err = np.array([thin_prism.trace(wave) - (nbk7(wave) - 1) * wedge_angle for wave in test_wavelengths])

    assert np.allclose(err, 0, atol=1e-3)

def test_prism():
    nbk7 = get_refractive_index("N-BK7")
    wedge_angle = np.deg2rad(3. + 53.0 / 60.0)
    prism = Prism(0, wedge_angle, nbk7)

    test_wavelengths = np.array([650.0, 700.0, 750.0]) * 1e-9
    deviation_angles = np.deg2rad(np.array([2.00390481945681, 1.9982250862238, 1.9934348068561]))
    err = np.array([prism.trace(wave) - dev for wave, dev in zip(test_wavelengths, deviation_angles)])

    assert np.allclose(np.rad2deg(err), 0, atol=1e-3)
    assert np.allclose(prism.minimal_deviation_angle(750e-9), 0.03471595671032256)

    prism.prism_angle = 2 * np.deg2rad(3. + 53.0 / 60.0)
    assert np.allclose(prism.minimal_deviation_angle(750e-9), 0.06958405951915544)

    Dtel = 1
    Dgrid = 1.1 * Dtel
    sr = np.mean(test_wavelengths) / Dtel

    grid = make_pupil_grid(128, Dgrid)
    aperture = make_circular_aperture(Dtel)(grid)

    focal_grid = make_focal_grid(q=3, num_airy=15, spatial_resolution=sr)
    prop = FraunhoferPropagator(grid, focal_grid)

    for wi, wave in enumerate(test_wavelengths):
        wf = Wavefront(aperture, wave)
        wf.total_power = 1

        prism.orientation = 2 * np.pi / 3 * wi
        prism_compensation = TiltElement(-prism.trace(wave), orientation=2 * np.pi / 3 * wi)
        wf_deviated = prop(prism_compensation(prism(wf)))
        wf_foc = prop(wf)

        assert np.max(abs(wf_deviated.power - wf_foc.power) / wf_foc.power.max()) < 1e-8



from .atmospheric_model import AtmosphericLayer
from ..util import inverse_tikhonov

class ModalAdaptiveOpticsLayer(AtmosphericLayer):
    def __init__(self, layer, controlled_modes, lag):
        self.layer = layer
        self.controlled_modes = controlled_modes

        AtmosphericLayer.__init__(self, layer.input_grid, layer.Cn_squared, layer.L0, layer.velocity, layer.height)

        self.transformation_matrix = controlled_modes.transformation_matrix
        self.transformation_matrix_inverse = inverse_tikhonov(self.transformation_matrix, 1e-7)

        self.corrected_coeffs = []
        self.lag = lag

    def phase_for(self, wavelength):
        ps = self.layer.phase_for(wavelength)
        ps -= self.transformation_matrix.dot(self.corrected_coeffs[0] / wavelength)

        return ps

    def evolve_until(self, t):
        self.layer.evolve_until(t)

        coeffs = self.transformation_matrix_inverse.dot(self.layer.phase_for(1))
        if len(self.corrected_coeffs) > self.lag:
            self.corrected_coeffs.pop(0)
        self.corrected_coeffs.append(coeffs)

    @property
    def Cn_squared(self):  # noqa: N802
        return self.layer.Cn_squared

    @Cn_squared.setter
    def Cn_squared(self, Cn_squared):  # noqa: N802
        self.layer.Cn_squared = Cn_squared

    @property
    def outer_scale(self):
        return self.layer.L0

    @outer_scale.setter
    def L0(self, L0):  # noqa: N802
        self.layer.L0 = L0

    def reset(self):
        self.corrected_coeffs = []
        self.layer.reset()



__all__ = [
    'make_fourier_transform',
    'FourierTransform',
    'ChirpZTransform',
    'make_fft_grid',
    'get_fft_parameters',
    'is_fft_grid',
    'FastFourierTransform',
    'FourierFilter',
    'MatrixFourierTransform',
    'NaiveFourierTransform',
    'ZoomFastFourierTransform',
]

from .fourier_transform import *
from .chirp_z_transform import *
from .fast_fourier_transform import *
from .fourier_operations import *
from .matrix_fourier_transform import *
from .naive_fourier_transform import *
from .zoom_fast_fourier_transform import *



from hcipy import *
import numpy as np
from math import *
import mpmath
import scipy

import pytest

def zernike_variance_von_karman(n, m, R, k0, Cn_squared, wavelength):
    '''Calculate the variance of the Zernike mode (`n`,`m`), using a von Karman turbulence spectrum.

    Parameters
    ----------
    n : int
        The radial Zernike order.
    m : int
        The azimuthal Zernike order.
    R : scalar
        The radius of the aperture.
    k0 : scalar
        The spatial frequency of the outer scale (1/L0).
    Cn_squared : scalar
        The integrated Cn^2 profile.
    wavelength : scalar
        The wavelength at which to calculate the variance.

    Returns
    -------
    scalar
        The variance of the specific Zernike mode.
    '''
    A = 0.00969 * (2 * np.pi / wavelength)**2 * Cn_squared
    coeffs_all = (-1)**(n - m) * 2 * (2 * np.pi)**(11 / 3) * (n + 1) * A * R**(5 / 3) / (sqrt(np.pi) * np.sin(np.pi * (n + 1 / 6)))

    term11 = mpmath.hyper([n + (3 / 2), n + 2, n + 1], [n + (1 / 6), n + 2, n + 2, 2 * n + 3], (2 * np.pi * R * k0)**2)
    term12 = sqrt(np.pi) * (2 * np.pi * R * k0)**(2 * n - 5 / 3) * scipy.special.gamma(n + 1) / (2**(2 * n + 3) * scipy.special.gamma(11 / 6) * scipy.special.gamma(n + 1 / 6) * scipy.special.gamma(n + 2)**2)
    term21 = -1 * scipy.special.gamma(7 / 3) * scipy.special.gamma(17 / 6) / (2 * scipy.special.gamma(-n + 11 / 6) * scipy.special.gamma(17 / 6)**2 * scipy.special.gamma(n + 23 / 6))
    term22 = mpmath.hyper([11 / 6, 7 / 3, 17 / 6], [-n + 11 / 6, 17 / 6, 17 / 6, n + 23 / 6], (2 * np.pi * R * k0)**2)

    return coeffs_all * (term11 * term12 + term21 * term22)

def check_total_variance(wavelength, D_tel, fried_parameter, outer_scale, propagate_phase_screen):
    velocity = 10.0  # meters/sec
    num_modes = 1000

    pupil_grid = make_pupil_grid(64, D_tel)
    aperture = make_circular_aperture(D_tel)(pupil_grid)

    Cn_squared = Cn_squared_from_fried_parameter(fried_parameter, wavelength)
    layer = InfiniteAtmosphericLayer(pupil_grid, Cn_squared, outer_scale, [velocity / np.sqrt(2), velocity / np.sqrt(2)], use_interpolation=False)

    num_iterations = 2000
    total_variance = []

    for it in range(num_iterations):
        layer.reset(make_independent_realization=True)
        if propagate_phase_screen:
            layer.t = np.sqrt(2) * D_tel / velocity

        phase = layer.phase_for(wavelength)
        total_variance.append(np.var(phase[aperture > 0]))

    variance_measured = np.mean(total_variance)

    variance_theory = 0
    for i in range(num_modes):
        n, m = noll_to_zernike(i + 2)
        variance_theory += zernike_variance_von_karman(n, m, D_tel / 2., 1. / outer_scale, layer.Cn_squared, wavelength)

    assert (variance_measured / variance_theory - 1) < 0.1

def check_zernike_variances(wavelength, D_tel, fried_parameter, outer_scale, propagate_phase_screen):
    velocity = 10.0  # meters/sec
    num_modes = 50

    pupil_grid = make_pupil_grid(128, D_tel)

    Cn_squared = Cn_squared_from_fried_parameter(fried_parameter, wavelength)
    layer = InfiniteAtmosphericLayer(pupil_grid, Cn_squared, outer_scale, [velocity / np.sqrt(2), velocity / np.sqrt(2)], use_interpolation=False)

    zernike_modes = make_zernike_basis(num_modes + 20, D_tel, pupil_grid, starting_mode=2, radial_cutoff=False)

    weights = evaluate_supersampled(make_circular_aperture(D_tel), pupil_grid, 32)
    zernike_modes = ModeBasis([z * np.sqrt(weights) for z in zernike_modes])

    transformation_matrix = zernike_modes.transformation_matrix
    projection_matrix = inverse_tikhonov(transformation_matrix, 1e-9)

    num_iterations = 300
    mode_coeffs = []

    for it in range(num_iterations):
        layer.reset(make_independent_realization=True)
        if propagate_phase_screen:
            layer.t = np.sqrt(2) * D_tel / velocity

        phase = layer.phase_for(wavelength)
        coeffs = projection_matrix.dot(phase * np.sqrt(weights))[:num_modes]
        mode_coeffs.append(coeffs)

    variances_simulated = np.var(mode_coeffs, axis=0)

    variances_theory = []
    for j in range(num_modes):
        n, m = noll_to_zernike(j + 2)
        variances_theory.append(zernike_variance_von_karman(n, m, D_tel / 2., 1. / outer_scale, layer.Cn_squared, wavelength))
    variances_theory = np.array(variances_theory)
    '''
    plt.plot(variances_simulated, label='simulated')
    plt.plot(variances_theory, label='theory')
    plt.yscale('log')
    plt.xlabel('Noll index')
    plt.ylabel('Variance (rad^2)')
    plt.legend()
    plt.show()
    '''
    assert np.all(np.abs(variances_simulated / variances_theory - 1) < 1)

def test_finite_atmosphere_total_variance():
    check_total_variance(0.5e-6, 1, 0.1, 10, False)

@pytest.mark.slow
def test_finite_atmosphere_total_variance_in_depth():
    # Check some selected parameter sets.
    check_total_variance(2e-6, 1, 0.1, 10, False)
    check_total_variance(0.5e-6, 8, 0.1, 10, False)
    check_total_variance(0.5e-6, 1, 0.3, 10, False)
    check_total_variance(0.5e-6, 1, 0.1, 40, False)

def test_infinite_atmosphere_total_variance():
    check_total_variance(0.5e-6, 1, 0.1, 10, True)

@pytest.mark.slow
def test_infinite_atmosphere_total_variance_in_depth():
    # Check some selected parameter sets.
    check_total_variance(2e-6, 1, 0.1, 10, True)
    check_total_variance(0.5e-6, 8, 0.1, 10, True)
    check_total_variance(0.5e-6, 1, 0.3, 10, True)
    check_total_variance(0.5e-6, 1, 0.1, 40, True)

def test_finite_atmosphere_zernike_variances():
    check_zernike_variances(0.5e-6, 1, 0.1, 10, False)

@pytest.mark.slow
def test_finite_atmosphere_zernike_variances_in_depth():
    # Check some selected parameter sets.
    check_zernike_variances(2e-6, 1, 0.1, 10, False)
    check_zernike_variances(0.5e-6, 8, 0.1, 10, False)
    check_zernike_variances(0.5e-6, 1, 0.3, 10, False)
    check_zernike_variances(0.5e-6, 1, 0.1, 40, False)

def test_infinite_atmosphere_zernike_variances():
    check_zernike_variances(0.5e-6, 1, 0.1, 10, True)

@pytest.mark.slow
def test_infinite_atmosphere_zernike_variances_in_depth():
    # Check some selected parameter sets.
    check_zernike_variances(2e-6, 1, 0.1, 10, True)
    check_zernike_variances(0.5e-6, 8, 0.1, 10, True)
    check_zernike_variances(0.5e-6, 1, 0.3, 10, True)
    check_zernike_variances(0.5e-6, 1, 0.1, 40, True)

@pytest.mark.parametrize('layer_cls', [InfiniteAtmosphericLayer, FiniteAtmosphericLayer])
def test_atmospheric_layer_reset(layer_cls):
    fried_parameter = 0.3  # meter
    wavelength = 1e-6  # meter
    velocity = 10.0  # meter / sec
    pupil_diameter = 8  # meter
    outer_scale = 20  # meter

    pupil_grid = make_pupil_grid(32, pupil_diameter)

    Cn_squared = Cn_squared_from_fried_parameter(fried_parameter, wavelength)
    layer = layer_cls(pupil_grid, Cn_squared, outer_scale, [velocity / np.sqrt(2), velocity / np.sqrt(2)])

    phase_1 = layer.phase_for(wavelength)
    layer.reset(make_independent_realization=False)
    phase_2 = layer.phase_for(wavelength)
    layer.reset(make_independent_realization=False)
    phase_3 = layer.phase_for(wavelength)
    layer.reset(make_independent_realization=True)
    phase_4 = layer.phase_for(wavelength)
    layer.reset(make_independent_realization=False)
    phase_5 = layer.phase_for(wavelength)
    layer.evolve_until(1)
    phase_6 = layer.phase_for(wavelength)
    layer.reset(make_independent_realization=False)
    layer.evolve_until(1)
    phase_7 = layer.phase_for(wavelength)

    assert np.allclose(phase_1, phase_2)
    assert np.allclose(phase_2, phase_3)
    assert not np.allclose(phase_3, phase_4)
    assert np.allclose(phase_4, phase_5)
    assert not np.allclose(phase_5, phase_6)
    assert np.allclose(phase_6, phase_7)

def test_multi_layer_atmosphere():
    r0 = 0.1
    wavelength = 500e-9

    pupil_grid = make_pupil_grid(256, 1.5)
    layers = make_standard_atmospheric_layers(pupil_grid)

    atmospheric_model = MultiLayerAtmosphere(layers, scintillation=False)
    atmospheric_model.Cn_squared = Cn_squared_from_fried_parameter(r0, wavelength)
    atmospheric_model.reset()

    aperture = Field(np.exp(-(pupil_grid.as_('polar').r / 0.65)**20), pupil_grid)
    wf = Wavefront(aperture, wavelength)

    # Model with no scintillation should only not modify amplitude
    wf_out = atmospheric_model.forward(wf)
    assert np.allclose(wf_out.amplitude, aperture)

    # It only should modify phase
    electric_field_compare = aperture * np.exp(1j * atmospheric_model.phase_for(wavelength))
    electric_field_simulated = wf_out.electric_field.copy()

    electric_field_compare /= electric_field_compare.mean()
    electric_field_simulated /= electric_field_simulated.mean()
    assert np.allclose(electric_field_compare, electric_field_simulated)

    wf_in = atmospheric_model.backward(wf_out)
    assert np.allclose(wf_in.electric_field, aperture)

    # Scintillation should also modify amplitude of the wavefront
    atmospheric_model.scintillation = True
    wf_out = atmospheric_model.forward(wf)
    assert not np.allclose(wf_out.amplitude, aperture)

    # Trying to get the phase for a model with scintillation should throw an exception
    with pytest.raises(ValueError):
        atmospheric_model.phase_for(wavelength)

    # Evolving a wavefront should modify the underlying layers
    atmospheric_model.evolve_until(1)
    wf_out2 = atmospheric_model.forward(wf)
    assert not np.allclose(wf_out.electric_field, wf_out2.electric_field)

def test_las_campanas_atmosphere():
    r0 = 0.1
    wavelength = 500e-9

    grid = make_pupil_grid(256, 1.5)
    layers = make_las_campanas_atmospheric_layers(grid)
    atmospheric_model = MultiLayerAtmosphere(layers, scintillation=False)
    atmospheric_model.Cn_squared = Cn_squared_from_fried_parameter(r0, wavelength)
    atmospheric_model.reset()

    aperture = Field(np.exp(-(grid.as_('polar').r / 0.65)**20), grid)
    wf = Wavefront(aperture, wavelength)

    # Model with no scintillation should only not modify amplitude
    wf_out = atmospheric_model.forward(wf)
    assert np.allclose(wf_out.amplitude, aperture)


def test_fried_parameter_seeing():
    for i in range(10):
        seeing = np.random.uniform(0.5, 2)  # arcsec
        wavelength = np.random.uniform(500e-9, 800e-9)

        fried_parameter = seeing_to_fried_parameter(seeing, wavelength)
        seeing_recovered = fried_parameter_to_seeing(fried_parameter, wavelength)

        assert np.allclose(seeing_recovered, seeing)

def test_phase_covariance_phase_structure_function_von_karman():
    grid = make_uniform_grid([256, 256], [1, 1], has_center=True)

    for r0 in [0.1, 0.2]:
        for L0 in [10, 20]:
            phase_structure_function = phase_structure_function_von_karman(r0, L0)(grid)
            phase_covariance = phase_covariance_von_karman(r0, L0)(grid)

            phase_structure_function_from_covariance = 2 * (phase_covariance.max() - phase_covariance)

            assert np.allclose(phase_structure_function_from_covariance, phase_structure_function)



from copy import copy
import numpy as np

from ..field import Field

def imshow_field(
        field, grid=None, ax=None, vmin=None, vmax=None, aspect='equal', norm=None, interpolation='nearest',
        non_linear_axes=False, cmap=None, mask=None, mask_color='k', grid_units=1, *args, **kwargs):
    '''Display a two-dimensional image on a matplotlib figure.

    This function serves as an easy replacement for the matplotlib.pyplot.imshow() function.
    Its signature mostly folows that of matplotlib, with a few minor differences.

    Parameters
    ----------
    field : Field or ndarray
        The field that we want to display. If this is an ndarray,
        then the parameter `grid` needs to be supplied. If the field
        is complex, then it will be automatically fed into :func:`complex_field_to_rgb`.
        If the field is a vector field with length 3 or 4, these will be
        interpreted as an RGB or RGBA field.
    grid : Grid or None
        If a grid is supplied, it will be used instead of the grid of `field`.
    ax : matplotlib axes
        The axes which to draw on. If it is not given, the current axes will be used.
    vmin : scalar
        The minimum value on the colorbar. If it is not given, then the minimum value
        of the field will be used.
    vmax : scalar
        The maximum value on the colorbar. If it is not given, then the maximum value
        of the field will be used.
    aspect : ['auto', 'equal', scalar]
        If 'auto', changes the image aspect ratio to match that of the axes.
        If 'equal', changes the axes aspect ratio to match that of the image.
    norm : Normalize
        A Normalize instance is used to scale the input to the (0, 1) range for
        input to the `cmap`. If it is not given, a linear scale will be used.
    interpolation : string
        The interpolation method used. The default is 'nearest'. Supported values
        are {'nearest', 'bilinear'}.
    non_linear_axes : boolean
        If axes are scaled in a non-linear way, for example on a log plot, then imshow_field
        needs to use a more expensive implementation. This parameter is to indicate that this
        algorithm needs to be used.
    cmap : Colormap or None
        The colormap with which to plot the image. It is ignored if a complex
        field or a vector field is supplied.
    mask : field or ndarray
        If part of the image needs to be masked, this mask is overlayed on top of the image.
        This is for example useful when plotting a phase pattern on a certain aperture, which
        has no meaning outside of the aperture. Masks can be partially translucent, and will
        be automatically scaled between (0, 1). Zero means invisible, one means visible.
    mask_color : Color
        The color of the mask, if it is used.
    grid_units : scalar or array_like
        The size of a unit square. The grid will be scaled by the inverse of this number before
        plotting. If this is a scalar, an isotropic scaling will be applied.

    Returns
    -------
    AxesImage
        The produced image.
    '''
    import matplotlib as mpl
    import matplotlib.pyplot as plt
    from matplotlib.image import NonUniformImage

    if ax is None:
        ax = plt.gca()

    ax.set_aspect(aspect)

    # Set/Find the correct grid and scale according to received grid units.
    if grid is None:
        if np.allclose(grid_units, 1):
            grid = field.grid
        else:
            grid = field.grid.scaled(1.0 / grid_units)
    else:
        if np.allclose(grid_units, 1):
            field = Field(field, grid)
        else:
            grid = grid.scaled(1.0 / grid_units)
            field = Field(field, grid)

    # If field is complex, draw complex
    if np.iscomplexobj(field):
        f = complex_field_to_rgb(field, rmin=vmin, rmax=vmax, norm=norm)
        vmin = None
        vmax = None
        norm = None
    else:
        f = field

    # Automatically determine vmin, vmax, norm if not overridden
    if norm is None and not np.iscomplexobj(field):
        if vmin is None:
            vmin = np.nanmin(f)
        if vmax is None:
            vmax = np.nanmax(f)
        norm = mpl.colors.Normalize(vmin, vmax)

    # Get extent
    c_grid = grid.as_('cartesian')
    min_x, min_y, max_x, max_y = c_grid.x.min(), c_grid.y.min(), c_grid.x.max(), c_grid.y.max()

    if grid.is_separated and grid.is_('cartesian'):
        # We can draw this directly
        x, y = grid.coords.separated_coords
        z = f.shaped
        if np.iscomplexobj(field) or field.tensor_order > 0:
            z = np.rollaxis(z, 0, z.ndim)
    else:
        # We can't draw this directly.
        raise NotImplementedError()

    if non_linear_axes:
        # Use pcolormesh to display
        x_mid = (x[1:] + x[:-1]) / 2
        y_mid = (y[1:] + y[:-1]) / 2

        x2 = np.concatenate(([1.5 * x[0] - 0.5 * x[1]], x_mid, [1.5 * x[-1] - 0.5 * x[-2]]))
        y2 = np.concatenate(([1.5 * y[0] - 0.5 * y[1]], y_mid, [1.5 * y[-1] - 0.5 * y[-2]]))
        X, Y = np.meshgrid(x2, y2)

        im = ax.pcolormesh(X, Y, z, *args, norm=norm, rasterized=True, cmap=cmap, **kwargs)
    else:
        # Use NonUniformImage to display
        im = NonUniformImage(ax, extent=(min_x, max_x, min_y, max_y), interpolation=interpolation, norm=norm, cmap=cmap, *args, **kwargs)
        im.set_data(x, y, z)

        from matplotlib.patches import Rectangle
        patch = Rectangle((min_x, min_y), max_x - min_x, max_y - min_y, facecolor='none')
        ax.add_patch(patch)
        im.set_clip_path(patch)

        ax.add_image(im)

    ax.set_xlim(min_x, max_x)
    ax.set_ylim(min_y, max_y)

    if mask is not None:
        one = np.ones(grid.size)
        col = mpl.colors.to_rgb(mask_color)

        m = np.array([one * col[0], one * col[1], one * col[2], 1 - mask / np.nanmax(mask)])

        imshow_field(m, grid, ax=ax)

    num_rows, num_cols = field.grid.shape
    def format_coord(x, y):  # pragma: no cover
        col = int(np.round((x - min_x) / (max_x - min_x) * (num_cols - 1)))
        row = int(np.round((y - min_y) / (max_y - min_y) * (num_rows - 1)))

        if col >= 0 and col < num_cols and row >= 0 and row < num_rows:
            z = field.shaped[row, col]
            if np.iscomplexobj(z):
                return 'x=%0.3g, y=%0.3g, z=%0.3g + 1j * %0.3g = %0.3g * exp(1j * %0.2f)' % (x, y, z.real, z.imag, np.abs(z), np.angle(z))
            else:
                return 'x=%0.3g, y=%0.3g, z=%0.3g' % (x, y, z)
        return 'x=%0.3g, y=%0.3g' % (x, y)

    ax.format_coord = format_coord

    ax._sci(im)

    return im

def imsave_field(filename, field, grid=None, vmin=None, vmax=None, norm=None, mask=None, mask_color='k', cmap=None):
    '''Save a two-dimensional field as an image.

    Parameters
    ----------
    filename : str
        The path to the file in which to save the field.
    field : Field or ndarray
        The field that we want to display. If this is an ndarray,
        then the parameter `grid` needs to be supplied. If the field
        is complex, then it will be automatically fed into :func:`complex_field_to_rgb`.
        If the field is a vector field with length 3 or 4, these will be
        interpreted as an RGB or RGBA field.
    grid : Grid or None
        If a grid is supplied, it will be used instead of the grid of `field`.
    vmin : scalar
        The minimum value on the colorbar. If it is not given, then the minimum value
        of the field will be used.
    vmax : scalar
        The maximum value on the colorbar. If it is not given, then the maximum value
        of the field will be used.
    norm : Normalize
        A Normalize instance is used to scale the input to the (0, 1) range for
        input to the `cmap`. If it is not given, a linear scale will be used.
    mask : field or ndarray
        If part of the image needs to be masked, this mask is overlayed on top of the image.
        This is for example useful when plotting a phase pattern on a certain aperture, which
        has no meaning outside of the aperture. Masks can be partially translucent, and will
        be automatically scaled between (0, 1). Zero means invisible, one means visible.
    mask_color : Color
        The color of the mask, if it is used.
    cmap : Colormap or None
        The colormap with which to plot the image. It is ignored if a complex
        field or a vector field is supplied.
    '''
    import matplotlib as mpl
    import matplotlib.pyplot as plt

    if grid is None:
        grid = field.grid
    else:
        field = Field(field, grid)

    # If field is complex, draw complex
    if np.iscomplexobj(field):
        f = complex_field_to_rgb(field, rmin=vmin, rmax=vmax, norm=norm)
        vmin = None
        vmax = None
        norm = None
    else:
        if norm is None:
            if vmin is None:
                vmin = np.nanmin(field)
            if vmax is None:
                vmax = np.nanmax(field)
            norm = mpl.colors.Normalize(vmin, vmax)
        f = field

    if mask is not None:
        f[~mask.astype('bool')] = np.nan

    try:
        cmap = mpl.colormaps.get_cmap(cmap)
    except AttributeError:
        # For Matplotlib <3.5.
        cmap = mpl.cm.get_cmap(cmap)

    cmap = copy(cmap)
    cmap.set_bad(mask_color)

    plt.imsave(filename, f.shaped, cmap=cmap, vmin=vmin, vmax=vmax)

def contour_field(field, grid=None, ax=None, grid_units=1, *args, **kwargs):
    '''Plot contours of a field.

    Parameters
    ----------
    field : Field or ndarray
        The field that we want to display. If this is an ndarray,
        then the parameter `grid` needs to be supplied. If the field
        is complex, then it will be automatically fed into :func:`complex_field_to_rgb`.
        If the field is a vector field with length 3 or 4, these will be
        interpreted as an RGB or RGBA field.
    grid : Grid or None
        If a grid is supplied, it will be used instead of the grid of `field`.
    ax : matplotlib axes
        The axes which to draw on. If it is not given, the current axes will be used.
    grid_units : scalar or array_like
        The size of a unit square. The grid will be scaled by the inverse of this number before
        plotting. If this is a scalar, an isotropic scaling will be applied.

    Returns
    -------
    QuadContourSet
        The plotted contour set.
    '''
    import matplotlib.pyplot as plt
    if ax is None:
        ax = plt.gca()

    # Set/Find the correct grid and scale according to received grid units.
    if grid is None:
        if np.allclose(grid_units, 1):
            grid = field.grid
        else:
            grid = field.grid.scaled(1.0 / grid_units)
    else:
        if np.allclose(grid_units, 1):
            field = Field(field, grid)
        else:
            grid = grid.scaled(1.0 / grid_units)
            field = Field(field, grid)

    c_grid = grid.as_('cartesian')
    min_x, min_y, max_x, max_y = c_grid.x.min(), c_grid.y.min(), c_grid.x.max(), c_grid.y.max()

    if grid.is_separated and grid.is_('cartesian'):
        # We can contour directly
        x, y = grid.coords.separated_coords
        z = field.shaped

        X, Y = np.meshgrid(x, y)

        cs = ax.contour(X, Y, z, *args, **kwargs)
    else:
        raise NotImplementedError()

    ax.set_xlim(min_x, max_x)
    ax.set_ylim(min_y, max_y)

    return cs

def contourf_field(field, grid=None, ax=None, grid_units=1, *args, **kwargs):
    '''Plot filled contours of a field.

    Parameters
    ----------
    field : Field or ndarray
        The field that we want to display. If this is an ndarray,
        then the parameter `grid` needs to be supplied. If the field
        is complex, then it will be automatically fed into :func:`complex_field_to_rgb`.
        If the field is a vector field with length 3 or 4, these will be
        interpreted as an RGB or RGBA field.
    grid : Grid or None
        If a grid is supplied, it will be used instead of the grid of `field`.
    ax : matplotlib axes
        The axes which to draw on. If it is not given, the current axes will be used.
    grid_units : scalar or array_like
        The size of a unit square. The grid will be scaled by the inverse of this number before
        plotting. If this is a scalar, an isotropic scaling will be applied.

    Returns
    -------
    QuadContourSet
        The plotted contour set.
    '''
    import matplotlib.pyplot as plt
    if ax is None:
        ax = plt.gca()

    # Set/Find the correct grid and scale according to received grid units.
    if grid is None:
        if np.allclose(grid_units, 1):
            grid = field.grid
        else:
            grid = field.grid.scaled(1.0 / grid_units)
    else:
        if np.allclose(grid_units, 1):
            field = Field(field, grid)
        else:
            grid = grid.scaled(1.0 / grid_units)
            field = Field(field, grid)

    c_grid = grid.as_('cartesian')
    min_x, min_y, max_x, max_y = c_grid.x.min(), c_grid.y.min(), c_grid.x.max(), c_grid.y.max()

    if grid.is_separated and grid.is_('cartesian'):
        # We can contour directly
        x, y = grid.coords.separated_coords
        z = field.shaped

        X, Y = np.meshgrid(x, y)

        cs = ax.contourf(X, Y, z, *args, **kwargs)
    else:
        raise NotImplementedError()

    ax.set_xlim(min_x, max_x)
    ax.set_ylim(min_y, max_y)

    return cs

def complex_field_to_rgb(field, theme='dark', rmin=None, rmax=None, norm=None):
    '''Convert a complex field to an RGB field.

    This function takes a scalar Field with complex numbers and converts it to
    a vector Field containing the red, green and blue values corresponding to the
    phase and amplitude of the complex input Field. The phase is decoded as hue, and
    amplitude as saturation/value. This function is especially useful for plotting complex
    fields.

    Parameters
    ----------
    field : scalar Field or ndarray
        The field that we want to convert.
    theme : ['dark', 'light']
        Whether to modulate saturation or value as a function of amplitude of the
        input field. Dark mode modulates value, light mode modulates saturation.
    rmin : scalar
        The minimum value on the colorbar. If this is not given, then the minimum
        of the absolute value of the Field will be used.
    rmax : scalar
        The maximum value on the colorbar. If this is not given, then the maximum
        of the absolute value of the Field will be used.
    norm : Normalize
        A Normalize instance is used to scale the absolute value to the (0, 1) range
        used for the value or saturation for the returned color. If it is not given,
        a linear scale will be used.

    Returns
    -------
    Field
        A vector field containing the red, green and blue components.

    Raises
    ------
    ValueError
        If the supplied field is not a scalar field.
    '''
    import matplotlib as mpl

    if not field.is_scalar_field:
        raise ValueError('Field must be a scalar field.')

    if norm is None:
        if rmin is None:
            rmin = np.nanmin(np.abs(field))
        if rmax is None:
            rmax = np.nanmax(np.abs(field))
        norm = mpl.colors.Normalize(rmin, rmax, True)

    hsv = np.zeros((field.size, 3), dtype='float')
    hsv[..., 0] = np.angle(field) / (2 * np.pi) % 1

    t = norm(np.abs(field))
    if theme == 'light':
        hsv[..., 1] = t
        hsv[..., 2] = 1
    elif theme == 'dark':
        hsv[..., 1] = 1
        hsv[..., 2] = t

    rgb = mpl.colors.hsv_to_rgb(hsv)
    alpha = np.isfinite(field)[:, np.newaxis]

    res = np.concatenate((rgb, alpha), axis=1)

    return Field(res.T, field.grid)
